// InSocket.cpp
// released 2006-09-25
/*
Copyright (C) 2006  Anders Hedstrom (grymse@alhem.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include <Parse.h>
#include <arpa/telnet.h>
#include <Utility.h>

#include "InSocket.h"
#include "cstring.h"
#include "FutureHandler.h"
#include "Login.h"
#include "Player.h"
#include "Command.h"

#define  TELOPT_MXP        '\x5B'

static	const unsigned char will_mxp_str  [] = { IAC, WILL, TELOPT_MXP, '\0' };
static	const unsigned char start_mxp_str [] = { IAC, SB, TELOPT_MXP, IAC, SE, '\0' };
static	const unsigned char do_mxp_str    [] = { IAC, DO, TELOPT_MXP, '\0' };
static	const unsigned char dont_mxp_str  [] = { IAC, DONT, TELOPT_MXP, '\0' };


InSocket::InSocket(SocketHandler& h)
:TcpSocket(h)
,IEventOwner(static_cast<FutureHandler&>(h))
,m_active_command(NULL)
,m_command_state(0)
,m_b_use_mxp(false)
,m_move_event(0)
{
	SetLineProtocol();
}


InSocket::~InSocket()
{
	printf("~InSocket()\n");
}


void InSocket::OnDelete()
{
	Player pl(GetDatabase(), GetAccountName());
	static_cast<FutureHandler&>(Handler()).ClearEvents(this);
	if (pl.Exists())
	{
		static_cast<FutureHandler&>(Handler()).GlobalEvent(this, pl.GetDisplayName() + " dropped.");
	}
}


void InSocket::OnAccept()
{
	static_cast<FutureHandler&>(Handler()).Count("Number of connections");
	printf("%s\n", GetRemoteAddress().c_str());
	Send("&WWelcome &rto &gFuture&nMud.\n");
	write_to_buffer(will_mxp_str);
	SetPrompt(static_cast<FutureHandler&>(Handler()).GetLogin(), 0, "Name, please? ");
}


void InSocket::Send(const std::string& s)
{
	cstring cs = s;
//	TcpSocket::Send(cs.c_str());
	write_to_buffer(cs.c_str());
}


void InSocket::Sendf(char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	char slask[5000];
#ifdef _WIN32
	vsprintf(slask, format, ap);
#else
	vsnprintf(slask, 5000, format, ap);
#endif
	va_end(ap);
	Send( slask );
}


void InSocket::SetPrompt(const std::string& p,bool colorize)
{
	m_active_command = NULL;
	if (!p.size()) // default
	{
		m_prompt = "> ";
	}
	else
	{
		m_prompt = p;
	}
	m_colorize_prompt = colorize;
	SendPrompt();
}


void InSocket::SetPrompt(Command *cc,int state,const std::string& prompt,bool colorize)
{
	m_active_command = cc;
	m_command_state = state;
	m_prompt = prompt;
	m_colorize_prompt = colorize;
	SendPrompt();
}


void InSocket::SendPrompt()
{
	if (!m_colorize_prompt)
	{
		write_to_buffer(m_prompt.c_str());
	}
	else
	{
		Send(m_prompt);
	}
}


void InSocket::ReadLine()
{
	if (ibuf.GetLength())
	{
		size_t x = 0;
		size_t n = ibuf.GetLength();
		char tmp[TCP_BUFSIZE_READ + 1];

		n = (n >= TCP_BUFSIZE_READ) ? TCP_BUFSIZE_READ : n;
		ibuf.Read(tmp,n);
		tmp[n] = 0;

		for (size_t i = 0; i < n; i++)
		{
			unsigned char uc;
			memcpy(&uc, &tmp[i], 1);
			if (uc == IAC)
			{
				size_t l = strlen((char *)do_mxp_str);
				if (!memcmp(tmp + i, do_mxp_str, l))
				{
					m_b_use_mxp = true;
					send_mxp_defines();
					memmove(tmp + i, tmp + i + l, strlen(tmp + i + l) + 1);
					n -= l;
				}
				else
				if (!memcmp(tmp + i, dont_mxp_str, strlen((char *)dont_mxp_str)))
				{
					m_b_use_mxp = false;
					l = strlen((char *)dont_mxp_str);
					memmove(tmp + i, tmp + i + l, strlen(tmp + i + l) + 1);
					n -= l;
				}
			}
			if (i >= n)
				break;
			while (tmp[i] == 13 || tmp[i] == 10)
			{
				char c = tmp[i];
				tmp[i] = 0;
				if (tmp[x])
				{
					m_line += (tmp + x);
				}
				OnLine( m_line );
				i++;
				if (i < n && (tmp[i] == 13 || tmp[i] == 10) && tmp[i] != c)
				{
					i++;
				}
				x = i;
				m_line = "";
			}
		}
		if (tmp[x])
		{
			m_line += (tmp + x);
		}
	}
}


void InSocket::OnLine(const std::string& line)
{
	if (line[0] == ' ')
	{
		size_t i = 0;
		while (line[i] == ' ')
			i++;
		static_cast<FutureHandler&>(Handler()).Talk(this, line.substr(i));
		SendPrompt();
		return;
	}
	if (m_active_command)
	{
		m_active_command -> OnLine(this, line, m_command_state);
		return;
	}
	if (!line.size())
	{
		SetPrompt();
		return;
	}
	Parse pa(line);
	std::string cmd = pa.getword();
	static_cast<FutureHandler&>(Handler()).DoCommand(this, cmd, pa.getrest());
}


void InSocket::send_mxp_defines() // http://www.gammon.com.au/mushclient/addingservermxp.htm
{
 	write_to_buffer( start_mxp_str );
	write_to_buffer( MXPMODE (6) );	 /* permanent secure mode */
	write_to_buffer( MXPTAG ("!-- Set up MXP elements --"));
	/* Exit tag */
	write_to_buffer( MXPTAG ("!ELEMENT Ex '<send>' FLAG=RoomExit"));
	/* Room description tag */
	write_to_buffer( MXPTAG ("!ELEMENT rdesc '<p>' FLAG=RoomDesc"));
	/* Get an item tag (for things on the ground) */
	write_to_buffer( MXPTAG 
			("!ELEMENT Get \"<send href='"
					 "get &#39;&name;&#39;|"
					 "examine &#39;&name;&#39;|"
					 "drink &#39;&name;&#39;"
			 "' "
			 "hint='RH mouse click to use this object|"
					 "Get &desc;|"
					 "Examine &desc;|"
					 "Drink from &desc;"
			 "'>\" ATT='name desc'") );
	/* Drop an item tag (for things in the inventory) */
	write_to_buffer( MXPTAG 
			("!ELEMENT Drop \"<send href='"
					 "drop &#39;&name;&#39;|"
					 "examine &#39;&name;&#39;|"
					 "look in &#39;&name;&#39;|"
					 "wear &#39;&name;&#39;|"
					 "eat &#39;&name;&#39;|"
					 "drink &#39;&name;&#39;"
			 "' "
			 "hint='RH mouse click to use this object|"
					 "Drop &desc;|"
					 "Examine &desc;|"
					 "Look inside &desc;|"
					 "Wear &desc;|"
					 "Eat &desc;|"
					 "Drink &desc;"
			 "'>\" ATT='name desc'") );
	/* Bid an item tag (for things in the auction) */
	write_to_buffer( MXPTAG 
			("!ELEMENT Bid \"<send href='bid &#39;&name;&#39;' "
			 "hint='Bid for &desc;'>\" "
			 "ATT='name desc'") );
	/* List an item tag (for things in a shop) */
	write_to_buffer( MXPTAG 
			("!ELEMENT List \"<send href='buy &#39;&name;&#39;' "
			 "hint='Buy &desc;'>\" "
			 "ATT='name desc'") );
	/* Player tag (for who lists, tells etc.) */
	write_to_buffer( MXPTAG 
			("!ELEMENT Player \"<send href='tell &#39;&name;&#39; ' "
			 "hint='Send a message to &name;' prompt>\" "
			 "ATT='name'") );
}


void InSocket::write_to_buffer(const char *s)
{
	std::string tmp;
	size_t l = strlen(s);
	bool in_mxp = false; // MXP_BEGc ... MXP_ENDc
	bool in_ent = false; // MXP_AMPc ... ;
	for (size_t i = 0; i < l; i++)
	{
		if (in_mxp)
		{
			if (s[i] == MXP_ENDc)
			{
				if (m_b_use_mxp)
					tmp += '>';
				in_mxp = false;
			}
			else
			{
				if (m_b_use_mxp)
					tmp += s[i];
			}
		}
		else
		if (in_ent)
		{
			if (m_b_use_mxp)
				tmp += s[i];
			if (s[i] == ';')
				in_ent = false;
		}
		else
		switch (s[i])
		{
		case MXP_BEGc:
			if (m_b_use_mxp)
				tmp += '<';
			in_mxp = true;
			break;
		case MXP_AMPc:
			if (m_b_use_mxp)
				tmp += '&';
			in_ent = true;
			break;
		case MXP_ENDc: // huh
			break;
		case '<':
			if (m_b_use_mxp)
				tmp += "&lt;";
			else
				tmp += s[i];
			break;
		case '>':
			if (m_b_use_mxp)
				tmp += "&gt;";
			else
				tmp += s[i];
			break;
		case '&':
			if (m_b_use_mxp)
				tmp += "&amp;";
			else
				tmp += s[i];
			break;
		case '"':
			if (m_b_use_mxp)
				tmp += "&quot;";
			else
				tmp += s[i];
			break;
		default:
			tmp += s[i];
		}
	}
	TcpSocket::Send(tmp);
}


void InSocket::write_to_buffer(const unsigned char *s)
{
	SendBuf((const char *)s, strlen((char *)s));
}


Database& InSocket::GetDatabase()
{
	return static_cast<FutureHandler&>(Handler()).GetDatabase();
}


void InSocket::OnEvent(int id)
{
	if (id == m_move_event)
	{
		m_move_event = 0;
		if (m_move_buffer.size())
		{
			std::list<std::string>::iterator it = m_move_buffer.begin();
			std::string move = *it;
			m_move_buffer.erase(it);
			Command *cc = NULL;
			if (move == "forward")
				cc = static_cast<FutureHandler&>(Handler()).GetCommand("forward");
			else
			if (move == "enter")
				cc = static_cast<FutureHandler&>(Handler()).GetCommand("enter");
			else
			switch (move[0])
			{
			case 'n':
				cc = static_cast<FutureHandler&>(Handler()).GetCommand("north");
				break;
			case 's':
				cc = static_cast<FutureHandler&>(Handler()).GetCommand("south");
				break;
			case 'e':
				cc = static_cast<FutureHandler&>(Handler()).GetCommand("east");
				break;
			case 'w':
				cc = static_cast<FutureHandler&>(Handler()).GetCommand("west");
				break;
			case 'u':
				cc = static_cast<FutureHandler&>(Handler()).GetCommand("up");
				break;
			case 'd':
				cc = static_cast<FutureHandler&>(Handler()).GetCommand("down");
				break;
			}
			if (cc)
			{
				std::string tmp = "nomap";
				Parse pa(tmp);
				if (move == "forward")
				{
					Send("\nMoving forward\n");
				}
				else
				if (move == "enter")
				{
					Send("\nEntering portal...\n");
				}
				else
				{
					Send("\nMoving " + cc -> GetCommand() + "\n");
				}
				cc -> Execute(this, tmp, pa);
			}
		}
	}
	else
	{
		Send("\nEvent: " + Utility::l2string(id) + "\n");
		SendPrompt();
	}
}


int InSocket::AddEvent(long sec,long usec)
{
	return static_cast<FutureHandler&>(Handler()).AddEvent(this, sec, usec);
}


void InSocket::BufferMove(const std::string& x)
{
	m_move_buffer.push_back(x);
	Send("Buffering move.\n");
	SetPrompt();
}


void InSocket::SetMoveEvent(long sec,long usec)
{
	m_move_event = AddEvent(sec, usec);
}


void InSocket::CancelMoves()
{
	while (m_move_buffer.size())
	{
		std::list<std::string>::iterator it = m_move_buffer.begin();
		m_move_buffer.erase(it);
	}
}


void InSocket::GlobalEvent(const std::string& x)
{
	static_cast<FutureHandler&>(Handler()).GlobalEvent(this, x);
}


void InSocket::Count(const std::string& x)
{
	static_cast<FutureHandler&>(Handler()).Count(x);
}


